/*
 * Ali M560x driver
 *
 * Copyright (C) 2005 m560x Linux Driver Project <m560x@x3ng.com.br>
 * Code to usb interface and ALI driver sofwtare:
 * copyright (c) 2006 Willem Duinker
 * Code for v4l2 interface copied in part from ov511 and sn9C102 drivers
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 * This driver is based on the 2.6.3 version of drivers/usb/usb-skeleton.c
 * but has been rewritten to be easy to read and use, as no locks are now
 * needed anymore.
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/kref.h>
#include <linux/version.h>
#include <asm/uaccess.h>
#include <linux/videodev.h>
#include <linux/usb.h>
#include <linux/firmware.h>



#include "m560x.h"
#define M560X_FIRMWARE_FILENAME "m560x.bin"
#define M560X_INF_FILENAME "m560x.inf"
#define FIRMWARE_STEP_SIZE 0x3e8
#define HASHSIZE 199

/* module parameter */
static unsigned int debug = 0;

/* table of usb-devices that work with this driver */
static struct usb_device_id m560x_table [] = {
	{ USB_DEVICE(USB_M560X_VENDOR_ID, USB_M560X_PRODUCT_M5602) },
	{ USB_DEVICE(USB_M560X_VENDOR_ID, USB_M560X_PRODUCT_M5603c) },
	{ }					/* Terminating entry */
};
MODULE_DEVICE_TABLE (usb, m560x_table);

                              
static void *rvmalloc(unsigned long size)
{
	void *mem;
	unsigned long adr;

	size = PAGE_ALIGN(size);
	mem = vmalloc_32(size);
	if (!mem)
		return NULL;

	memset(mem, 0, size); /* Clear the ram out, no junk to the user */
	adr = (unsigned long) mem;
	while (size > 0) {
		SetPageReserved(vmalloc_to_page((void *)adr));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}

	return mem;
}

static void rvfree(void *mem, unsigned long size)
{
	unsigned long adr;

	if (!mem)
		return;

	adr = (unsigned long) mem;
	while ((long) size > 0) {
		ClearPageReserved(vmalloc_to_page((void *)adr));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
	vfree(mem);
}

static int read_ali_register(struct m560x_camera *cam, __u16 ali_register) {
  struct usb_device* udev = cam->udev;
  u8* buff = cam->control_buffer;
  int result = 0;

  result = usb_control_msg(udev, usb_rcvctrlpipe(udev, 0), 0x04, 0xc0,0x14, ali_register, buff, 1, 0x100);
//   PDEBUG(DBG_TRACE,"register %04x result %02x value %02x",ali_register,result,*buff);
  return (result >= 0) ? (int)(*buff) : -1;
}

int write_ali_register(struct m560x_camera *cam, __u16 ali_register, unsigned char value) {
  struct usb_device* udev = cam->udev;
  u8* buff = cam->control_buffer;
  int result = 0;
  
  buff[0] = value;
  
  result = usb_control_msg(udev, usb_sndctrlpipe(udev, 0),0x04,0x40,0x13,ali_register,buff,1,100);
//   PDEBUG(DBG_TRACE,"register %04x result %02x value %02x",ali_register,result,*buff);
  return (result >= 0) ? (int)(*buff) : -1;
}

int or_ali_register(struct m560x_camera *cam, __u16 ali_register, unsigned char mask) {
  struct usb_device* udev = cam->udev;
  u8* buff = cam->control_buffer;
  int result;

  result = usb_control_msg(udev, usb_rcvctrlpipe(udev, 0), 0x04, 0xc0, 0x0014, ali_register, buff, 1, 0x100);
//   PDEBUG(DBG_TRACE,"register %04x result %02x value %02x",ali_register,result,*buff);
  if ( result != 1 ) {
    return -1;
  }
  buff[0] |= mask;
//   PDEBUG(DBG_TRACE,"mask %02x data to return %02x",mask,buff[0]);
  result = usb_control_msg(udev, usb_sndctrlpipe(udev, 0), 0x04, 0x40,0x0013,ali_register,buff,1,100);
//   PDEBUG(DBG_TRACE,"register %04x result %02x value %02x",ali_register,result,*buff);
  return (result >= 0) ? (int)(*buff) : -1;
}

int and_ali_register(struct m560x_camera *cam, __u16 ali_register, __u8 mask) {
  struct usb_device* udev = cam->udev;
  u8* buff = cam->control_buffer;
  int result;

  result = usb_control_msg(udev, usb_rcvctrlpipe(udev, 0), 0x04, 0xc0, 0x0014, ali_register, buff, 1, 0x100);
  //PDEBUG(DBG_TRACE,"register %04x result %02x value %02x",ali_register,result,*buff);
  if ( result != 1 ) {
    return -1;
  }
  buff[0] &= mask;
 // PDEBUG(DBG_TRACE,"mask %02x data to return %02x",mask,buff[0]);
  result = usb_control_msg(udev, usb_sndctrlpipe(udev, 0), 0x04, 0x40,0x0013,ali_register,buff,1,100);
  //PDEBUG(DBG_TRACE,"register %04x result %02x value %02x",ali_register,result,*buff);
  return (result >= 0) ? (int)(*buff) : -1;
}


int write_sensor_register(struct m560x_camera *cam, __u16 ali_register, unsigned char value) {
  struct usb_device* udev = cam->udev;
  u8* buff = cam->control_buffer;
  int result = 0;
  
  buff[0] = value;
  
  result = usb_control_msg(udev, usb_sndctrlpipe(udev, 0),0x04,0x40,0x17,ali_register,buff,1,100);
//   PDEBUG(DBG_TRACE,"register %04x result %02x value %02x",ali_register,result,*buff);
  return (result >= 0) ? (int)(*buff) : -1;
}

int halt_dsp(struct m560x_camera *cam) {
  u8* buff = cam->control_buffer;
  int result = 0;

  result = read_ali_register(cam,0x8000);
  PDEBUG(DBG_TRACE,"0x8000 %02x",result);
  result = write_ali_register(cam,0x8000,0x40);
  PDEBUG(DBG_TRACE,"0x8000 %02x",result);
  result = read_ali_register(cam,0x8140);
  PDEBUG(DBG_TRACE,"0x8140 %02x",result);
  
  return (result >= 0) ? (int)(*buff) : -1;
}

int restart_dsp(struct m560x_camera *cam) {
  struct usb_device* udev = cam->udev;
  u8* buff = cam->control_buffer;
  int result = 0;

  buff[0] = 0x00;
  result = usb_control_msg(udev, usb_sndctrlpipe(udev, 0),0x04,0x40,0x10,0x0000,buff,1,100);
  PDEBUG(DBG_TRACE,"register %04x result %02x value %02x",0x0000,result,*buff);
  return (result >= 0) ? (int)(*buff) : -1;
}

static int getTokens(int * tokens, char * line, int max) {
  int i = 0;
  int count = 0;
  int in_string = 0;

   do {
     if ( line[i] == 0x00) {
        break;
     }
    else if ( line[i] == 0x09 || line[i] == 0x20 || line[i] == 0x2c || line[i] == 0x3b || line[i] == 0x0d) {
      in_string = 0;
      line[i] = '\0';
    }
    else if ( in_string == 0) {
        in_string = 1;
        tokens[count] = i;
        count++;
    }

    i++;
   } while (count < (max + 1));
  
   return count;
}

static unsigned hash(char *s)
{
  unsigned hashval;

  for (hashval = 0; * s!= '\0'; s++)
    hashval = *s + 31 * hashval;
  return hashval % HASHSIZE;
}


static struct nlist *lookup(char *s, struct nlist **hashtab)
{
  struct nlist *np;

  for (np = hashtab[hash(s)];np != NULL;np = np->next)
    if (strcmp(s, np->name) == 0)
      return np;
  return NULL;
}


int init_from_script(struct m560x_camera *cam, char * pitem) {
  const struct firmware *firmware;
  void *ptr1, *ptr2;
  char * line;
  int i;
  int size,length;
  int retval;
  unsigned char type_id;
  int result = 0;
  int tokens[20];
  struct nlist * np;
  unsigned char value;
  u16 reg_value;

  PDEBUG(DBG_TRACE,"init from script %s",pitem);
  retval = request_firmware(&firmware, M560X_SCRIPT_FILENAME, &cam->udev->dev);
  line = ( char *) cam->control_buffer;

  ptr1 = firmware->data;
  size = firmware->size;

  // find appropriate part of script

  do {
    ptr2 = memscan(ptr1,0x0a,size);
    length = ptr2 - ptr1;
    for ( i = 0; i < length; i++) {
      line[i] = *((char *) ptr1);
      ptr1++;
    }
    line[length] = 0x00;
    type_id = line[0];
    ptr1 = ptr2 + 1;
    ptr2 = firmware->data;
    if ( (ptr1 >= ptr2 + size)) {
      release_firmware(firmware);
//       PDEBUG(DBG_TRACE,"End of File reached");
      return -1;
    }
  }
  while (!(( type_id == '[' ) && ( strstr(&line[1],pitem) != NULL)));

  // process appropriate part of script
  do {
    ptr2 = memscan(ptr1,0x0a,size);
    length = ptr2 - ptr1;
    for ( i = 0; i < length; i++) {
      line[i] = *((char *) ptr1);
      ptr1++;
    }
    line[length] = 0x00;
//     PDEBUG(DBG_TRACE,"line %s",line);
    type_id = line[0];
    switch (type_id) {
      case 'R':
      case 'O':
      case 'A':
        getTokens(tokens,line,3);
        np = lookup(&line[tokens[1]],cam->hashtab);
        if ( np == NULL )
          return -1;
        value = (unsigned char) simple_strtoul(&line[tokens[2]],NULL,16);
//       PDEBUG(DBG_TRACE,"%c %s %04x %02x\t",type_id,&line[tokens[1]],np->entry_value,value);
        switch (type_id) {
          case 'R':
//             PDEBUG(DBG_TRACE,"register write %04x %02x\t",np->entry_value,value);
            write_ali_register(cam,np->entry_value,value);
            break;
          case 'O':
//             PDEBUG(DBG_TRACE,"or register %04x %02x\t",np->entry_value,value);
            or_ali_register(cam,np->entry_value,value);
            break;
          case 'A':
//             PDEBUG(DBG_TRACE,"and register %04x %02x\t",np->entry_value,value);
            and_ali_register(cam,np->entry_value,value);
            break;
          case 'S':
//             PDEBUG(DBG_TRACE,"read register %04x %02x\t",np->entry_value,value);
            read_ali_register(cam,np->entry_value);
            break;
          default:
            break;
        }
        break;
      case 'I':
        getTokens(tokens,line,4);
        reg_value = (unsigned int) ((simple_strtoul(&line[tokens[2]],NULL,16) << 8) | simple_strtoul(&line[tokens[1]],NULL,16));
        value = (unsigned char) simple_strtoul(&line[tokens[3]],NULL,16);  // for kernel simple_strtoul
//         PDEBUG(DBG_TRACE,"%c %04x %02x\t",type_id,reg_value,value);
        write_sensor_register(cam,reg_value,value);
        break;
      default:
        break;
    }
    ptr1 = ptr2 + 1;
    ptr2 = firmware->data;
    if ( (ptr1 >= ptr2 + size))
      break;
  }
  while ( type_id != '[');

  release_firmware(firmware);
  return result;
}

int load_firmware(struct m560x_camera *cam)
{
  const struct firmware *firmware;
  int retval;
  unsigned int offset = 0;
  int count = 0;
  int size, i;
  unsigned char * buff, * fwbuf, *pbuf, *rbuf;
  
  struct usb_device* udev = cam->udev;
  
  buff = kmalloc(0x40, GFP_KERNEL);
  retval = request_firmware(&firmware, M560X_FIRMWARE_FILENAME,
                             &cam->udev->dev);
  PDEBUG(DBG_TRACE,"size of firmware file %08x",firmware->size);

  size = firmware->size;
  fwbuf = kmalloc(size, GFP_KERNEL);
  for ( i = 0; i < size; i++ ) {
    fwbuf[i] = firmware->data[i];
    if ( i < 5) 
      PDEBUG(DBG_TRACE,"test result %02x ",fwbuf[i]);
  }
  pbuf = fwbuf;

  while ( (size -= FIRMWARE_STEP_SIZE) > 0) {
    count = FIRMWARE_STEP_SIZE;
    PDEBUG(DBG_TRACE,"do max count %04x %d\n", offset,count);
    retval = usb_control_msg(cam->udev,usb_sndctrlpipe(udev, 0),0x04,0x40,0x15,offset,pbuf,count,100);
    PDEBUG(DBG_TRACE,"result %04x\n", retval);    
    offset += FIRMWARE_STEP_SIZE;
    pbuf += FIRMWARE_STEP_SIZE;
  }
  count = size + FIRMWARE_STEP_SIZE;
  PDEBUG(DBG_TRACE,"count %04x %d\n", offset,count);

  retval = usb_control_msg(cam->udev,usb_sndctrlpipe(udev, 0),0x04,0x40,0x0015,offset,pbuf,count,100);
  PDEBUG(DBG_TRACE,"result %d\n", retval);
  
  rbuf = buff;
  retval = usb_control_msg(cam->udev,usb_rcvctrlpipe(udev, 0),0x04,0xc0,0x0016,offset,buff,0x10,100);
  PDEBUG(DBG_TRACE,"result %d\n", retval);
  for ( i = 0; i < 5; i++) {
    PDEBUG(DBG_TRACE,"read result %02x %02x ",*(rbuf++),fwbuf[offset + i]);
    
  }
  
  release_firmware(firmware);
  kfree(buff);
  kfree(fwbuf);
  
  return (retval > 0 ? 0 : -1);
}

char * strdup(char *name) {
  char * copy;
    
  copy = kmalloc(strlen(name), GFP_KERNEL);
  if (copy == NULL) {
    err("Out of memory");
    return NULL;
  }
  strcpy(copy,name);
  
  return copy;
}

struct nlist *install(char *name, unsigned short entry_value, struct nlist **hashtab)
{
  struct nlist *np;
  unsigned hashval;

//   PDEBUG(DBG_TRACE,"install %s %04x",name,entry_value);
  if ((np = lookup(name,hashtab)) == NULL) {
    np = (struct nlist *) kmalloc(sizeof(*np), GFP_KERNEL);
    if (np == NULL || (np->name = strdup(name)) == NULL)
      return NULL;
    hashval = hash(name);
    np->next = hashtab[hashval];
    hashtab[hashval] = np;
  }
  np->entry_value = entry_value;
  return np;
}

void clear_hashtab(struct nlist **hashtab)
{
  int i;
  struct nlist *np;
  int empty = 0;
  
  for ( i = 0; i < HASHSIZE; i++ ) {
    np = hashtab[i];
    if ( np == NULL)
      empty++;
//     PDEBUG(DBG_TRACE,"hashindex %d",i);
    for (np = hashtab[i];np != NULL;np = np->next) {
      if ( np->name != NULL) {
//         PDEBUG(DBG_TRACE,"clear entry %s",np->name);
        kfree((void *) np->name);
      }
      kfree(np);
    }
  }
  PDEBUG(DBG_TRACE,"empty nodes %d",empty);
}



static int init_hashtab(struct m560x_camera *cam) {
  const struct firmware *firmware;
  int retval;
  void *ptr1, *ptr2;
  char * line;
  int i,n;
  unsigned int  entry_value;
  int tokens[20];
  int size, length;
  unsigned char type_id;
  struct nlist **hashtab = cam->hashtab;

  PDEBUG(DBG_TRACE,"Initialize hashtab with ali register names");
  retval = request_firmware(&firmware, M560X_INF_FILENAME,
                             &cam->udev->dev);

  PDEBUG(DBG_TRACE,"Size of script file %d",firmware->size);

  line = ( char *) cam->control_buffer;

  ptr1 = firmware->data;
  size = firmware->size;

  // find appropriate part of script

  do {
    ptr2 = memscan(ptr1,0x0a,size);
    length = ptr2 - ptr1;
    for ( i = 0; i < length; i++) {
      line[i] = *((char *) ptr1);
      ptr1++;
    }
    line[length] = 0x00;
    type_id = line[0];
    ptr1 = ptr2 + 1;
    ptr2 = firmware->data;
    if ( (ptr1 >= ptr2 + size)) {
      release_firmware(firmware);
      err("End of File reached");
      return -1;
    }
  }
  while (!(( type_id == '[' ) && ( strstr(&line[1],"SensorRegSetting") != NULL)));

  do {
    ptr2 = memscan(ptr1,0x0a,size);
    length = ptr2 - ptr1;
    for ( i = 0; i <= length + 1; i++) {
      line[i] = *((char *) ptr1);
      ptr1++;
    }
    line[length] = 0x00;
    type_id = line[0];
    n = getTokens(tokens,line,15);
    if ( n > 1) {
      entry_value =  (unsigned int) simple_strtoul(&line[tokens[n-1]],NULL,16);
      install(&line[tokens[2]],entry_value,hashtab);
    }
    ptr1 = ptr2 + 1;
    ptr2 = firmware->data;
    if ( (ptr1 >= ptr2 + size)) {
      release_firmware(firmware);
      err("End of File reached");
      return -1;
    }
  }
  while ( type_id != '[');

  release_firmware(firmware);

  return 0;
}

static void m560x_delete(struct kref *kref)
{
//   struct m560x_camera *cam = to_m560x_camera(kref);

 /* if (cam) {                                 // resource release in disconnect
  usb_put_dev(cam->udev);
  if (cam->int_in_buffer)
  kfree (cam->int_in_buffer);
  if (cam->isoc_in_buffer)
  kfree (cam->isoc_in_buffer);
  cam->udev = NULL;
      
  kfree (cam);
}*/
}

static int initialize_camera(struct m560x_camera *cam)
{
  int result = 0;
  
  PDEBUG(DBG_TRACE, "initialize_camera");
  
  halt_dsp(cam);
  load_firmware(cam);
  init_hashtab(cam);
  restart_dsp(cam);
  result = init_from_script(cam,"DSP_INIT");
  if ( result < 0)  {
    err("error from init_from_script");
    return -1;
  }
  init_from_script(cam,"AP_ON");
  init_from_script(cam,"R_VGA");
  init_from_script(cam,"EXTRA2");
	
  return 0;
}

static int set_camera_power(struct m560x_camera *cam, int state)
{
	PDEBUG(DBG_TRACE, "set_camera_power");

	info("set_camera_power - not implemented yet...");

	return 0;
}

/* Bytes per frame. Used by read(). Return of 0 indicates error */
static inline long int
    get_frame_length(struct m560x_frame *frame)
{
  if (!frame)
    return 0;
  else 
    return (3 *frame->width * frame->height);
}

typedef enum
{
  BAYER_PATTERN_BGGR,
  BAYER_PATTERN_GRBG,
  BAYER_PATTERN_RGGB,
  BAYER_PATTERN_GBRG
} bayer_pattern_t;

void
    BayerNearestNeighbor( unsigned char *src, 
                          unsigned char *dest, 
                          int sx, 
                          int sy, 
                          bayer_pattern_t type)
{
  unsigned char *outR, *outG, *outB;
  register int i,j;
 // sx and sy should be even
  switch (type) {
    case BAYER_PATTERN_GRBG:
    case BAYER_PATTERN_BGGR:
      outR=&dest[0];
      outG=&dest[1];
      outB=&dest[2];
      break;
    case BAYER_PATTERN_GBRG:
    case BAYER_PATTERN_RGGB:
/*      outR=&dest[2];
      outG=&dest[1];
      outB=&dest[0];*/
      outB=&dest[2];
      outG=&dest[1];
      outR=&dest[0];
      break;
    default:
      outR=NULL;outG=NULL;outB=NULL;
      break;
  }
  
  switch (type) {
    case BAYER_PATTERN_GRBG: //-------------------------------------------
    case BAYER_PATTERN_GBRG:
   // copy original RGB data to output images
      for (i=0;i<sy;i+=2) {
        for (j=0;j<sx;j+=2) {
          outG[(i*sx+j)*3]=src[i*sx+j];
          outG[((i+1)*sx+(j+1))*3]=src[(i+1)*sx+(j+1)];
          outR[(i*sx+j+1)*3]=src[i*sx+j+1];
          outB[((i+1)*sx+j)*3]=src[(i+1)*sx+j];
      //    printf("%d %d %d %d %08x %08x\n",i,j,((i+1)*sx+j)*3,(i+1)*sx+(j+1),((i+1)*sx+j)*3,(i+1)*sx+(j+1));
        }
      }
     // R channel
      for (i=0;i<sy;i+=2) {
        for (j=0;j<sx-1;j+=2) {
          outR[(i*sx+j)*3]=outR[(i*sx+j+1)*3];
          outR[((i+1)*sx+j+1)*3]=outR[(i*sx+j+1)*3];
          outR[((i+1)*sx+j)*3]=outR[(i*sx+j+1)*3];
        }
      }
      // B channel
      for (i=0;i<sy-1;i+=2)  { //every two lines
        for (j=0;j<sx-1;j+=2) {
          outB[(i*sx+j)*3]=outB[((i+1)*sx+j)*3];
          outB[(i*sx+j+1)*3]=outB[((i+1)*sx+j)*3];
          outB[((i+1)*sx+j+1)*3]=outB[((i+1)*sx+j)*3];
        }
      }
    // using lower direction for G channel
      
    // G channel
      for (i=0;i<sy-1;i+=2)//every two lines
        for (j=1;j<sx;j+=2)
          outG[(i*sx+j)*3]=outG[((i+1)*sx+j)*3];
      
      for (i=1;i<sy-2;i+=2)//every two lines
        for (j=0;j<sx-1;j+=2)
          outG[(i*sx+j)*3]=outG[((i+1)*sx+j)*3];
    
    // copy it for the next line
      for (j=0;j<sx-1;j+=2)
        outG[((sy-1)*sx+j)*3]=outG[((sy-2)*sx+j)*3];
    
      break;
      case BAYER_PATTERN_BGGR: //-------------------------------------------
    case BAYER_PATTERN_RGGB:
    // copy original data
      for (i=0;i<sy;i+=2) {
        for (j=0;j<sx;j+=2) {
          outB[(i*sx+j)*3]=src[i*sx+j];
          outR[((i+1)*sx+(j+1))*3]=src[(i+1)*sx+(j+1)];
          outG[(i*sx+j+1)*3]=src[i*sx+j+1];
          outG[((i+1)*sx+j)*3]=src[(i+1)*sx+j];
        }
      }
    // R channel
      for (i=0;i<sy;i+=2){
        for (j=0;j<sx-1;j+=2) {
          outR[(i*sx+j)*3]=outR[((i+1)*sx+j+1)*3];
          outR[(i*sx+j+1)*3]=outR[((i+1)*sx+j+1)*3];
          outR[((i+1)*sx+j)*3]=outR[((i+1)*sx+j+1)*3];
        }
      }
    // B channel
      for (i=0;i<sy-1;i+=2) { //every two lines
        for (j=0;j<sx-1;j+=2) {
          outB[((i+1)*sx+j)*3]=outB[(i*sx+j)*3];
          outB[(i*sx+j+1)*3]=outB[(i*sx+j)*3];
          outB[((i+1)*sx+j+1)*3]=outB[(i*sx+j)*3];
        }
      }
    // using lower direction for G channel
    
    // G channel
      for (i=0;i<sy-1;i+=2)//every two lines
        for (j=0;j<sx-1;j+=2)
          outG[(i*sx+j)*3]=outG[((i+1)*sx+j)*3];
    
      for (i=1;i<sy-2;i+=2)//every two lines
        for (j=0;j<sx-1;j+=2)
          outG[(i*sx+j+1)*3]=outG[((i+1)*sx+j+1)*3];
    
    // copy it for the next line
      for (j=0;j<sx-1;j+=2)
        outG[((sy-1)*sx+j+1)*3]=outG[((sy-2)*sx+j+1)*3];
    
      break;
    
      default:  //-------------------------------------------
        break;
  }
}

static void m560x_postprocess(struct m560x_camera *cam, struct m560x_frame *frame)
{
  int frmx;
  int width,height;
  PDEBUG(DBG_TRACE,"Bayer Nearest Neighbor frame %d",cam->grabframe);
      
  frmx = cam->grabframe;
  width = 640;
  height = 480;
  BayerNearestNeighbor( frame->rawdata, frame->data, width, height, (bayer_pattern_t) BAYER_PATTERN_RGGB);
}

static int m560x_new_frame(struct m560x_camera *cam, int framenum)
{
  struct m560x_frame *frame;
  int newnum;

  PDEBUG(DBG_TRACE, "ov->curframe = %d, framenum = %d", cam->curframe, framenum);

  if (!cam->udev)
    return -1;

  /* If we're not grabbing a frame right now and the other frame is */
  /* ready to be grabbed into, then use it instead */
  if (cam->curframe == -1) {
    newnum = (framenum - 1 + M560X_NUMFRAMES) % M560X_NUMFRAMES;
    if (cam->frame[newnum].grabstate == FRAME_READY)
      framenum = newnum;
  } else
    return 0;

    frame = &cam->frame[framenum];

    PDEBUG(4, "framenum = %d, width = %d, height = %d", framenum,
           frame->width, frame->height);

    frame->grabstate = FRAME_GRABBING;
    frame->scanstate = STATE_SCANNING;
    
    cam->curframe = framenum;

    return 0;
}

// static inline void m560x_move_data(struct m560x_camera *cam, unsigned char *in, int n)
// {
//   struct m560x_frame *frame = &cam->frame[cam->curframe];
//   
//   int nextf;
//   int offset;
//   int num;
// 
// 
//       frame->grabstate = FRAME_DONE;
//       wake_up_interruptible(&frame->wq);
// 
//       /* If next frame is ready or grabbing,
//       * point to it */
//       nextf = (cam->curframe + 1) % M560X_NUMFRAMES;
//       if (cam->frame[nextf].grabstate == FRAME_READY
//           || cam->frame[nextf].grabstate == FRAME_GRABBING) {
//         cam->curframe = nextf;
//         cam->frame[nextf].scanstate = STATE_SCANNING;
//           } else {
//             if (frame->grabstate == FRAME_DONE) {
//               PDEBUG(4, "** Frame done **");
//             } else {
//               PDEBUG(4, "Frame not ready? state = %d",
//                      cam->frame[nextf].grabstate);
//             }
// 
//             cam->curframe = -1;
//           }
// 
//     /* Image corruption caused by misplaced frame->segment = 0
//     * fixed by carlosf@conectiva.com.br
//     */
// 
//     /* Frame start */
//           PDEBUG(DBG_TRACE, "Frame start, framenum = %d", cam->curframe);
//     frame->scanstate = STATE_LINES;
//     frame->bytes_recvd = 0;
// 
//     /* Are we in a frame? */
//     if (frame->scanstate != STATE_LINES) {
//   PDEBUG(DBG_TRACE, "Not in a frame; packet skipped");
//   return;
//     }
// 
//     /* If frame start, skip header */
//     if (frame->bytes_recvd == 0)
//       offset = 9;
//     else
//       offset = 0;
// 
//     num = n - offset - 1;
// 
// 
// }

static void isoc_urb_complete(struct urb * urb, struct pt_regs * regs) {
  struct m560x_camera * cam = urb->context;
  int i;
  int frmx;
  struct m560x_frame * frame;
  
  
//   PDEBUG(DBG_TRACE,"urb completion routine");
  for ( i = 0; i < urb->number_of_packets; i++) {
    unsigned int length, status,marker;
    unsigned char prev,next;
    int j;
    unsigned char * ptr;
    int detect, index;
    int end_var;
    
    frmx = cam->grabframe;
    detect = 0;
    index = 4;
     
    length = urb->iso_frame_desc[i].actual_length;
    status = urb->iso_frame_desc[i].status;
//    goto end;
//    ptr = urb->transfer_buffer + urb->iso_frame_desc[i].offset;
 //   for ( j = 0; j < 8; j++ ) {
//       PDEBUG(DBG_TRACE,"%02x",*ptr);
 //     ptr++;
 //   }
//    marker = 0;
//    ptr = urb->transfer_buffer + urb->iso_frame_desc[i].offset;
    if (length > 4) {
//    for ( j = 0; j < 4; j++ ) {
//      marker |= (((unsigned int) *ptr) << (3-j)*8);
 //     ptr++;
//       PDEBUG(DBG_TRACE,"%08x",marker);
//    }
//     PDEBUG(DBG_TRACE,"packet_nr %d marker %08x length %04x status %d",i,marker,length,status);
//    if ( marker != 0xffffffff )
//      PDEBUG(DBG_TRACE,"wrong packet marker");
//        PDEBUG(DBG_TRACE,"packet_nr %d length %04x status %d %08x",i,length,status,cam->readCount);
 //     ptr = urb->transfer_buffer + 4;
   ptr = urb->transfer_buffer + urb->iso_frame_desc[i].offset + 4;
      prev = 0x00;
    for ( j = 5; j <= length; j++) {
      next = *(ptr++);
//       PDEBUG(DBG_TRACE,"%02x",next);
      if ( (prev == 0xff) && (next == 0xff)) {
        cam->frameCount++;
        PDEBUG(DBG_TRACE,"found frame marker %d %d %08x",j,cam->frameCount,cam->readCount);
        detect = 1;
        index = j;
      }
      prev = next;
    }
    if ( frmx == -1 ) {
      frmx = 0;
      cam->frame[frmx].grabstate = FRAME_GRABBING;
    }
    cam->grabframe = frmx;
    ptr = urb->transfer_buffer + urb->iso_frame_desc[i].offset;
    frame = &cam->frame[frmx];
    if ( detect == 0) {
      if ( (cam->frame[frmx].bytes_recvd + length - 4) > (640*480)) {
        PDEBUG(DBG_TRACE,"exceed frame %d size %08x %08x",frmx,cam->frame[frmx].bytes_recvd,length);
        err("exceed frame size");
        return;
      }
//       PDEBUG(DBG_TRACE,"frame %d %08x %08x size %08x %08x",frmx,cam->frame[frmx].rawdata + cam->frame[frmx].bytes_recvd,ptr + 4,length -4,cam->frame[frmx].bytes_recvd);
      memcpy(cam->frame[frmx].rawdata + cam->frame[frmx].bytes_recvd,ptr + 4,length - 4);
      cam->frame[frmx].bytes_recvd += length - 4;
      cam->readCount += length -4;
//       PDEBUG(DBG_TRACE,"recvd %08x reaadCount %08x",cam->frame[frmx].bytes_recvd,cam->readCount);
    }
    else {
      if ( cam->frame[frmx].bytes_recvd > (640*480)) {
        PDEBUG(DBG_TRACE,"exceed frame size %08x",cam->frame[frmx].bytes_recvd);
        err("exceed frame size");
        return;
      }
      PDEBUG(DBG_TRACE,"frame change %d  %08x %08x size %08x %08x",frmx,cam->frame[frmx].rawdata + cam->frame[frmx].bytes_recvd,ptr + 4,index - 6,cam->frame[frmx].bytes_recvd);
      if ( index > 6) {
        PDEBUG(DBG_TRACE,"memcpy rest");
        memcpy(cam->frame[frmx].rawdata + cam->frame[frmx].bytes_recvd,ptr + 4,index - 6);
      }
      cam->frame[frmx].bytes_recvd += index - 6;
      cam->readCount += index - 6;
      PDEBUG(DBG_TRACE,"recvd %08x reaadCount %08x",cam->frame[frmx].bytes_recvd,cam->readCount);
      m560x_postprocess(cam,&cam->frame[frmx]);
      cam->frame[frmx].grabstate = FRAME_DONE;
      wake_up_interruptible(&cam->frame[frmx].wq);
      PDEBUG(DBG_TRACE,"finalize old frame %d",frmx);
      PDEBUG(DBG_TRACE,"found frame marker %d final size %08x",cam->frameCount,cam->readCount);
      frmx = (frmx + 1) % M560X_NUMFRAMES;
      cam->grabframe = frmx;
      cam->frame[frmx].bytes_recvd = 0;
      cam->frame[frmx].grabstate = FRAME_GRABBING;
      PDEBUG(DBG_TRACE,"new frame %d",frmx);
      ptr = urb->transfer_buffer + urb->iso_frame_desc[i].offset;
      PDEBUG(DBG_TRACE,"frame change %d  %08x %08x size %08x %08x",frmx,cam->frame[frmx].rawdata + cam->frame[frmx].bytes_recvd,ptr + index,length - index,cam->frame[frmx].bytes_recvd);
      memcpy(cam->frame[frmx].rawdata + cam->frame[frmx].bytes_recvd,ptr + index,length - index);
      cam->frame[frmx].bytes_recvd += length - index;
      cam->readCount = length -index;
      PDEBUG(DBG_TRACE,"recvd %08x reaadCount %08x",cam->frame[frmx].bytes_recvd,cam->readCount);
    }
 end:   
     end_var = 0; 
    }
    else {
      PDEBUG(DBG_TRACE,"empty packet length %04x status %d",length,status);
    }

  }
// /*  cam->packetCount++;
//   if ( cam->packetCount < 200) {
    usb_submit_urb(urb, GFP_ATOMIC);
//   }*/
  
}

static int create_buffers(struct m560x_camera * cam)
{
  int i;
  int retval = 0;
  const int width = 640;
  const int height = 480;
  const int data_bufsize = 3 * M560X_NUMFRAMES * width * height;
  const int raw_bufsize = M560X_NUMFRAMES * width * height;
  
  down(&cam->buf_lock);
  cam->framebuffers = rvmalloc(data_bufsize);
  cam->databuffers = vmalloc(raw_bufsize);
  
  for ( i = 0; i < M560X_NUMFRAMES; i++) {
    cam->frame[i].data = cam->framebuffers + i * 3 * width * height;
    cam->frame[i].rawdata= cam->databuffers + i * width * height;
  }
  up(&cam->buf_lock);
  
  return retval;
}

static int release_buffers(struct m560x_camera * cam)
{
  int i;
  int retval = 0;
  const int width = 640;
  const int height = 480;
  const int data_bufsize = 3 * M560X_NUMFRAMES * width * height;
  const int raw_bufsize = M560X_NUMFRAMES * width * height;

  down(&cam->buf_lock);
  
  if ( cam->framebuffers) {
    rvfree(cam->framebuffers,data_bufsize);
    cam->framebuffers = NULL;
  }
  if ( cam->databuffers) {
    rvfree(cam->databuffers,raw_bufsize);
    cam->databuffers = NULL;
  }
  
  for ( i = 0; i < M560X_NUMFRAMES; i++) {
    cam->frame[i].data = NULL;
    cam->frame[i].rawdata= NULL;
  }
  up(&cam->buf_lock); 
  
  return retval;
}

static int init_data(struct m560x_camera *cam)
{
  int retval = 0;
  int i,j;
  struct urb * urb;
  struct usb_device* udev = cam->udev;
  
  PDEBUG(DBG_TRACE,"Initialize isochronous buffers and urb's");
  PDEBUG(DBG_TRACE,"number of urb's %d packetsize %04x number of packets %d",NB_ISOC_URBS,ISOC_PKTSIZE,NB_ISOC_PACKETS);
  cam->isoc_in_size = ISOC_PKTSIZE;
  for (i = 0; i < NB_ISOC_URBS; i++) {
    cam->isoc_in_buffer[i] = kmalloc(NB_ISOC_PACKETS*ISOC_PKTSIZE, GFP_KERNEL);
    if ( !cam->isoc_in_buffer[i]) {
      retval = -ENOMEM;
      err("no memory for isoc buffer");
    }
  }
  
  for (i = 0; i < NB_ISOC_URBS; i++) {
    urb = usb_alloc_urb(NB_ISOC_PACKETS, GFP_KERNEL);
    if ( !urb) {
      retval = -ENOMEM;
      err("no memory for isoc buffer");
    } 
    cam->urb[i] = urb;
    urb->dev = udev;
    urb->context = cam;
    urb->pipe = usb_rcvisocpipe(udev, 1);
    urb->transfer_flags = URB_ISO_ASAP;
    urb->number_of_packets = NB_ISOC_PACKETS;
    urb->complete = isoc_urb_complete;
    urb->transfer_buffer = cam->isoc_in_buffer[i];
    urb->transfer_buffer_length = ISOC_PKTSIZE * NB_ISOC_PACKETS;
    urb->interval = 1;
    for (j = 0; j < NB_ISOC_PACKETS; j++) {
      urb->iso_frame_desc[j].offset = ISOC_PKTSIZE * j;
      urb->iso_frame_desc[j].length = ISOC_PKTSIZE;
      urb->iso_frame_desc[j].actual_length = 0;
      urb->iso_frame_desc[j].status = 0;
    }
  }
  
  usb_set_interface(udev,0,1);

/*  
  init_from_script(cam,"NEXT");
  init_from_script(cam,"NEXT2");*/
  
  for (i = 0; i < NB_ISOC_URBS; i++) {
  if ( cam->urb[i]) {
   retval = usb_submit_urb(cam->urb[i], GFP_KERNEL);
   if ( retval) {
     err("submit of urb failed %d",retval);
    
   }
  }
}
//   }

  return retval;
}

static int stop_data(struct m560x_camera *cam)
{
  int retval = 0;
  int i;
  
  PDEBUG(DBG_TRACE,"Free isochronous buffers and urb's");
  for ( i = NB_ISOC_URBS - 1; i >= 0; i--) {
    usb_kill_urb(cam->urb[i]);
    usb_free_urb(cam->urb[i]);
    kfree(cam->isoc_in_buffer[i]);
  }
  
  usb_set_interface(cam->udev,0,0);
  
  return retval;
}

/*
 *  ---------------------------------------------------------------------------
 * Video4Linux Functions
 * ---------------------------------------------------------------------------
 */
static int v4l_m560x_open(struct inode *inode, struct file *file)
{
	struct video_device *vdev = video_devdata(file);
	struct m560x_camera *cam =
	    (struct m560x_camera *) vdev->priv;
	int retval = 0;
 int i;

	PDEBUG(DBG_TRACE, "v4l_m560x_open");

	if (!cam) {
		err("m560x video_device improperly initialized");
		retval = -ENODEV;
		goto error;
	}

	/* increment our usage count for the device */
// 	kref_get(&cam->kref);

	/* save our object in the file's private structure */
	file->private_data = cam;

	// First upload firmware, then turn the camera on
// 	if (!cam->is_initialized) {
		initialize_camera(cam);

		cam->is_initialized = 1;
// 	}

	set_camera_power(cam, 1);
 
 /* Make sure frames are reset */
 for (i = 0; i < M560X_NUMFRAMES; i++) {
   cam->frame[i].grabstate = FRAME_UNUSED;
   cam->frame[i].bytes_read = 0;
   cam->frame[i].bytes_recvd = 0;
      cam->frame[i].framenum = i;
     init_waitqueue_head(&cam->frame[i].wq);
 }
 create_buffers(cam);

  init_data(cam);
  cam->packetCount = 0;
  cam->readCount = 0;
  cam->frameCount = 0;
	return 0;
	
error:
// 	if (cam)
// 		kref_put(&cam->kref, m560x_delete);

	return retval;
}

static int v4l_m560x_close(struct inode *inode, struct file *file)
{
	struct m560x_camera *cam = (struct m560x_camera *)file->private_data;
	int retval = 0;

	PDEBUG(DBG_TRACE, "v4l_m560x_close");

	if (!cam) {
		err("m560x video_device improperly initialized");
		retval = -ENODEV;
		goto error;
	}

	/* it's not the end of the world if
	 * we fail to turn the camera off.
	 */
  stop_data(cam);
  release_buffers(cam);
  set_camera_power(cam, 0);

	/* decrement the count on our device */
// 	kref_put(&cam->kref, m560x_delete);

	return 0;

error:
	return retval;
}

static ssize_t v4l_m560x_read(struct file *file, char __user *buffer, size_t acount, loff_t *ppos)
{
//   struct video_device *vdev = file->private_data;
  struct video_device *vdev = video_devdata(file);
//   struct sn9c102_device* cam = video_get_drvdata(video_devdata(filp));
  int noblock = file->f_flags & O_NONBLOCK;
  unsigned long count = acount;
  struct m560x_camera * cam = video_get_drvdata(vdev);
  struct m560x_frame * frame;
  int i;
  int retval = 0;
  int frmx = -1;

  PDEBUG(DBG_TRACE, "v4l_m560x_read");
  PDEBUG(DBG_TRACE, "file flags %08x %08x",file->f_flags,noblock);
/*  retval = 0x4b000;
  return retval;*/
  if (down_interruptible(&cam->lock))
  return -EINTR;
  
  if (!vdev || !buffer) {
    retval = -EFAULT;
    up(&cam->lock);
    return retval;
  }

  if(!cam->vdev) {
    retval = -ENODEV;
    up(&cam->lock);
    return retval;
  }
  
  // FIXME: Only supports two frames
  /* See if a frame is completed, then use it. */
  if (cam->frame[0].grabstate >= FRAME_DONE) /* _DONE or _ERROR */
    frmx = 0;
  else if (cam->frame[1].grabstate >= FRAME_DONE)/* _DONE or _ERROR */
    frmx = 1;
//   else
//     frmx = 0;

   PDEBUG(DBG_TRACE, "use frame %d %d",frmx,noblock);
  /* If nonblocking we return immediately */
  if (noblock && (frmx == -1)) {
    retval = -EAGAIN;
    up(&cam->lock);
    return retval;
  }
   PDEBUG(DBG_TRACE, "past noblocking");
  
  /* If no FRAME_DONE, look for a FRAME_GRABBING state. */
  /* See if a frame is in process (grabbing), then use it. */
  if (frmx == -1) {
    if (cam->frame[0].grabstate == FRAME_GRABBING)
      frmx = 0;
    else if (cam->frame[1].grabstate == FRAME_GRABBING)
      frmx = 1;
  }

  /* If no frame is active, start one. */
  if (frmx == -1) {
    if ((retval = m560x_new_frame(cam, frmx = 0))) {
      err("read: m560x_new_frame error");
      up(&cam->lock);
      return retval;
    }
  }

  frame = &cam->frame[frmx];

restart:
  if (!cam->udev) {
    retval = -EIO;
    up(&cam->lock);
    return retval;
  }

  /* Wait while we're grabbing the image */
  PDEBUG(DBG_TRACE, "Waiting image grabbing %d",frmx);
    retval = wait_event_interruptible(frame->wq,
                                  (frame->grabstate == FRAME_DONE)
                                      || (frame->grabstate == FRAME_ERROR));

  if (retval) {
    up(&cam->lock);
    return retval;
  }

  PDEBUG(DBG_TRACE, "Got image, frame->grabstate = %d", frame->grabstate);
  PDEBUG(DBG_TRACE, "bytes_recvd = %d", frame->bytes_recvd);

  if (frame->grabstate == FRAME_ERROR) {
      frame->bytes_read = 0;
      err("** ick! ** Errored frame %d", cam->curframe);
      if (m560x_new_frame(cam, frmx)) {
        err("read: m560x_new_frame error");
        up(&cam->lock);
        return retval;
      }
      goto restart;
  }

    /* Decompression, format conversion, etc... */
//     m560x_postprocess(cam, frame);

    PDEBUG(DBG_TRACE, "frmx=%d, bytes_read=%ld, length=%ld", frmx,
           frame->bytes_read,
           get_frame_length(frame));

    /* copy bytes to user space; we allow for partials reads */
//  if ((count + frame->bytes_read)
//      > get_frame_length((struct ov511_frame *)frame))
//    count = frame->scanlength - frame->bytes_read;

    /* FIXME - count hardwired to be one frame... */
    count = get_frame_length(frame);

    PDEBUG(DBG_TRACE, "Copy to user space: %ld bytes", count);
    if ((i = copy_to_user(buffer, frame->data + frame->bytes_read, count))) {
      PDEBUG(DBG_TRACE, "Copy failed! %d bytes not copied", i);
      retval = -EFAULT;
      up(&cam->lock);
      return retval;
    }

    frame->bytes_read += count;
    PDEBUG(DBG_TRACE, "{copy} count used=%ld, new bytes_read=%ld",
           count, frame->bytes_read);

    /* If all data have been read... */
    if (frame->bytes_read
        >= get_frame_length(frame)) {
      frame->bytes_read = 0;

// FIXME: Only supports two frames
      /* Mark it as available to be used again. */
      cam->frame[frmx].grabstate = FRAME_UNUSED;
      if ((retval = m560x_new_frame(cam, !frmx))) {
        err("m560x_new_frame returned error");
        up(&cam->lock);
        return retval;
      }
        }

        PDEBUG(DBG_TRACE, "read finished, returning %08x (sweet)", count);

        up(&cam->lock);
        return count;

}


static int v4l_m560x_mmap(struct file *file, struct vm_area_struct *vma)
{
	PDEBUG(DBG_TRACE, "v4l_m560x_mmap");
	info("v4l_m560x_mmap - not implemented yet...");
	return 0;
}


static int v4l_m560x_do_ioctl(struct inode *inode, struct file *file, unsigned int ioctlnr, unsigned long arg, v4l2_kioctl driver_ioctl)
{
  void __user *user_arg = (void __user *)arg;
  int retval = 0;
  struct m560x_camera *cam = (struct m560x_camera *)file->private_data;

  PDEBUG(DBG_TRACE, "v4l_m560x_ioctl %02x",(unsigned char) ioctlnr);

  if (!cam) {
    err("m560x video_device improperly initialized");
    retval = -ENODEV;
    goto exit;
  }

  switch (ioctlnr) {

    /* ------- tv norms --------- */
    case VIDIOC_ENUMSTD:
    {
      struct v4l2_standard *e = arg;

      PDEBUG(DBG_TRACE, "VIDIOC_ENUMSTD %d",e->index);
      if (e->index)
        retval = -EINVAL;
//       e->id = V4L2_STD_UNKNOWN;
      e->id = V4L2_STD_PAL_BG;
      return retval;
    }

    case VIDIOC_G_STD:
    {
      struct v4l2_standard *e = arg;
      
      PDEBUG(DBG_TRACE, "GET STD %d",e->index);
      retval = -EINVAL;
      break;
    }

    case VIDIOC_S_STD:
    {
      struct v4l2_standard *e = arg;
      
      PDEBUG(DBG_TRACE, "SET STD %d",e->index);
      break;
    }
    
    case VIDIOC_QUERYSTD:
    {
      PDEBUG(DBG_TRACE, "OTHER STD");
      retval = -EINVAL;
      return retval;
    }

    /* ------- input switching --------- */
    case VIDIOC_ENUMINPUT:
    {
      struct v4l2_input *i = arg;

      PDEBUG(DBG_TRACE, "VIDIOC_ENUMINPUT %d",i->index);
      strlcpy(i->name,"USB", sizeof(i->name));
      i->type = V4L2_INPUT_TYPE_CAMERA;

      if (i->index)
        retval = -EINVAL;

      break;
    }

    case VIDIOC_G_INPUT:
    {
      struct v4l2_input *i = arg;
      
      PDEBUG(DBG_TRACE, "GET INPUT %d",i->index);
      if (i->index)
        retval = -EINVAL;
      break;
    }
    case VIDIOC_S_INPUT:
    {
      struct v4l2_input *i = arg;

      if (i->index != 0)
        retval = -EINVAL;

      break;
    }

    /* ------- streaming parameters --------- */
    case VIDIOC_G_PARM:
    {
      struct v4l2_streamparm *sp = arg;
      PDEBUG(DBG_TRACE, "VIDIOC_G_PARM %d",sp->type);
      if (sp->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
        retval = -EINVAL;

      sp->parm.capture.capability = 0;
      sp->parm.capture.capturemode = 0;
      sp->parm.capture.timeperframe.numerator = 1;
      sp->parm.capture.timeperframe.denominator = 30;
      sp->parm.capture.readbuffers = 2;
      sp->parm.capture.extendedmode = 0;
      break;
    }

    /* audio interface - we have none */
    case VIDIOC_G_AUDIO:
    {
      PDEBUG(DBG_TRACE, "Get Audio");
      retval = -EINVAL;
      break;
    }
    case VIDIOC_S_AUDIO:
      retval = -EINVAL;
      break;

    /* ------- input switching --------- */
   // There are no parameters yet to control
    case VIDIOC_QUERYCTRL:
    {
      struct v4l2_queryctrl *q = arg;
      PDEBUG(DBG_TRACE, "VIDIOC_QUERYCTRL %08x",q->id);
      retval = -EINVAL;
     
      break;
    }
   
    case VIDIOC_S_TUNER:
    {
      struct v4l2_tuner *t = arg;
      
      PDEBUG(DBG_TRACE, "VIDIOC_G_PARM %d",t->index);
      retval = -EINVAL;
      break;
    }

    /* pointless to implement overlay with this camera */
    case VIDIOC_G_FBUF:
    case VIDIOC_S_FBUF:
    case VIDIOC_OVERLAY:
      retval = -EINVAL;
      break;

      /* tuner interface - we have none */
    case VIDIOC_G_TUNER:
    case VIDIOC_G_FREQUENCY:
    case VIDIOC_S_FREQUENCY:
      retval = -EINVAL;
      break;



    case VIDIOC_QUERYMENU:
      retval = -EINVAL;
      break;
    case VIDIOC_CROPCAP:
    {
      struct v4l2_cropcap cc ;

      cc.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      cc.pixelaspect.numerator = 1;
      cc.pixelaspect.denominator = 1;
      cc.bounds.top = 0;
      cc.bounds.left = 0;
      cc.bounds.width = 640;
      cc.bounds.height = 480;
      cc.defrect.top = 0;
      cc.defrect.left = 0;
      cc.defrect.width = 640;
      cc.defrect.height = 480;

      memcpy(user_arg, &cc, sizeof(cc));
//       if (copy_to_user(user_arg, &cc, sizeof(cc)))
//         return -EFAULT;

      return 0;
    }
   
    default:
     // return v4l_compat_translate_ioctl(inode,file,cmd,arg,
     //                                   video_do_ioctl);
      PDEBUG(DBG_TRACE, "translate %02x",(unsigned char) ioctlnr);
      retval =  v4l_compat_translate_ioctl(inode,file,ioctlnr,(void __user *)arg,(void *) driver_ioctl);
      PDEBUG(DBG_TRACE, "retval %08x",retval);
//       retval = -ENOIOCTLCMD;
      break;
  }

exit:
    return retval;
}

static int v4l_m560x_video_do_ioctl(struct inode *inode, struct file *filp,
                                    unsigned int cmd, void *arg)
{
  struct m560x_camera *cam = (struct m560x_camera *)filp->private_data;
  int retval = 0;

  PDEBUG(DBG_TRACE, "v4l_m560x_video_do_ioctl %02x",(unsigned char) cmd);

  if (!cam) {
    err("m560x video_device improperly initialized");
    retval = -ENODEV;
    return retval;
  }

  switch (cmd) {

    /* --- capabilities ------------------------------------------ */
    case VIDIOC_QUERYCAP:
    {
      struct v4l2_capability *cap = arg;

      PDEBUG(DBG_TRACE, "VIDIOC_QUERYCAP");
      memset(cap, 0, sizeof(*cap));
      strlcpy(cap->driver, "m560x", sizeof(cap->driver));
      cap->capabilities = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_READWRITE;
      cap->version = (__u32) DRIVER_VERSION_NUMERICAL,
      strlcpy(cap->card, cam->vdev->name, sizeof(cap->card));
      if (usb_make_path(cam->udev, cap->bus_info, sizeof(cap->bus_info)) < 0)
        strlcpy(cap->bus_info, cam->vdev->name, sizeof(cap->bus_info));
     break;
    }

    /* --- capture ioctls ------------------------------------------ */
    case VIDIOC_ENUM_FMT:
    {
      struct v4l2_fmtdesc *fmtd = arg;

      PDEBUG(DBG_TRACE, "VIDIOC_ENUM_FMT %d",fmtd->index);
      if (fmtd->index != 0)
        return -EINVAL;

      memset(fmtd,0,sizeof(*fmtd));
      fmtd->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      fmtd->flags = 0;
      strcpy(fmtd->description,"rgb24");
      fmtd->pixelformat = V4L2_PIX_FMT_RGB24;
      memset(fmtd->reserved,0,sizeof(fmtd->reserved));
      break;
    }
   
    case VIDIOC_G_FMT:
    {
      struct v4l2_format *fmtd = arg;
      struct v4l2_pix_format pix_format;

      PDEBUG(DBG_TRACE, "GET FMT %d",fmtd->type);
      if(fmtd->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
        retval = -EINVAL;
     
      pix_format.width = 640;
      pix_format.height = 480;
      pix_format.pixelformat = V4L2_PIX_FMT_RGB24;
      pix_format.field = V4L2_FIELD_NONE;
      pix_format.bytesperline = 3*pix_format.width;
      pix_format.sizeimage = pix_format.width * pix_format.height;
      pix_format.colorspace = V4L2_COLORSPACE_SRGB;
      pix_format.priv = 0;
     
      memcpy(&(fmtd->fmt.pix),&pix_format,sizeof(pix_format));
      
      break;
    }
   
    case VIDIOC_S_FMT:
    {
      struct v4l2_format *fmtd = arg;
      struct v4l2_pix_format pix_format;

      PDEBUG(DBG_TRACE, "SET FMT %d",fmtd->type);
      PDEBUG(DBG_TRACE, "SET FMT width %d",fmtd->fmt.pix.width);
      PDEBUG(DBG_TRACE, "SET FMT height %d",fmtd->fmt.pix.height);
      if (fmtd->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
        return -EINVAL;
      pix_format.width = 640;
      pix_format.height = 480;
      pix_format.pixelformat = V4L2_PIX_FMT_RGB24;
      pix_format.field = V4L2_FIELD_NONE;
      pix_format.bytesperline = 3*pix_format.width;
      pix_format.sizeimage = pix_format.width * pix_format.height;
      pix_format.colorspace = V4L2_COLORSPACE_SRGB;
      pix_format.priv = 0;

      memcpy(&(fmtd->fmt.pix),&pix_format,sizeof(pix_format));

      break;
    }

    case VIDIOC_TRY_FMT:
    {
      struct v4l2_format *fmtd = arg;
      struct v4l2_pix_format pix_format;

      PDEBUG(DBG_TRACE, "TRY FMT %d",fmtd->type);
      if (fmtd->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
        return -EINVAL;
      pix_format.width = 640;
      pix_format.height = 480;
      pix_format.pixelformat = V4L2_PIX_FMT_RGB24;
      pix_format.field = V4L2_FIELD_NONE;
      pix_format.bytesperline = 3*pix_format.width;
      pix_format.sizeimage = pix_format.width * pix_format.height;
      pix_format.colorspace = V4L2_COLORSPACE_SRGB;
      pix_format.priv = 0;

      memcpy(&(fmtd->fmt.pix),&pix_format,sizeof(pix_format));

      break;
    }

    default:
      return v4l_m560x_do_ioctl(inode, filp, cmd, arg,
                                v4l_m560x_video_do_ioctl);
  }
  return 0;
}

static int v4l_m560x_ioctl(struct inode *inode, struct file *file,
                           unsigned int cmd, unsigned long arg)
{
  return video_usercopy(inode, file, cmd, arg, v4l_m560x_video_do_ioctl);
}


static struct file_operations v4l_m560x_fops = {
	//.owner		= THIS_MODULE,
	.open		= v4l_m560x_open,
	.release	= v4l_m560x_close,
	.read		= v4l_m560x_read,
	.mmap		= v4l_m560x_mmap,
	.ioctl		= v4l_m560x_ioctl,
 .llseek		= no_llseek,
/*#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
     .compat_ioctl   = v4l_compat_ioctl32,
#endif*/
};


/****************************************************************************
 *  sysfs
 ***************************************************************************/
#define m560x_file(name, variable, field)                              \
static ssize_t show_##name(struct class_device *class_dev, char *buf)   \
{                                                                       \
        struct video_device *vdev = to_video_device(class_dev);         \
        struct m560x_camera *cam = video_get_drvdata(vdev);                  \
        return sprintf(buf, field, cam->variable);                      \
}                                                                       \
static CLASS_DEVICE_ATTR(name, S_IRUGO, show_##name, NULL);

m560x_file(model, name, "%s\n");
m560x_file(in_use, user, "%d\n");
m560x_file(streaming, streaming, "%d\n");
m560x_file(palette, palette, "%i\n");
m560x_file(frames_total, readCount, "%d\n");
m560x_file(frames_read, frameCount, "%d\n");
m560x_file(packets_dropped, dropped, "%d\n");
m560x_file(decoding_errors, error, "%d\n");

static void m560x_create_sysfs_files(struct video_device *vdev)
{
        video_device_create_file(vdev, &class_device_attr_model);
        video_device_create_file(vdev, &class_device_attr_in_use);
        video_device_create_file(vdev, &class_device_attr_streaming);
        video_device_create_file(vdev, &class_device_attr_palette);
        video_device_create_file(vdev, &class_device_attr_frames_total);
        video_device_create_file(vdev, &class_device_attr_frames_read);
        video_device_create_file(vdev, &class_device_attr_packets_dropped);
        video_device_create_file(vdev, &class_device_attr_decoding_errors);
}

static void m560x_remove_sysfs_files(struct video_device *vdev)
{
        video_device_remove_file(vdev, &class_device_attr_model);
        video_device_remove_file(vdev, &class_device_attr_in_use);
        video_device_remove_file(vdev, &class_device_attr_streaming);
        video_device_remove_file(vdev, &class_device_attr_palette);
        video_device_remove_file(vdev, &class_device_attr_frames_total);
        video_device_remove_file(vdev, &class_device_attr_frames_read);
        video_device_remove_file(vdev, &class_device_attr_packets_dropped);
        video_device_remove_file(vdev, &class_device_attr_decoding_errors);
}

/****************************************************************************
 * USB - Functions
 ***************************************************************************/
static int usb_m560x_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
  struct m560x_camera *cam = NULL;
  struct usb_device *dev = interface_to_usbdev(interface);
  struct usb_host_interface *iface_desc;
  struct usb_endpoint_descriptor *endpoint;
  char *camera_name = NULL;
  size_t buffer_size;
  int i;
  int retval = -ENOMEM;

  info("m560x_probe - M560x-based WebCam connected");

  /* allocate memory for our device state and initialize it */
  cam = kmalloc(sizeof(*cam), GFP_KERNEL);
  if (cam == NULL) {
    err("Out of memory");
    goto error;
  }
  memset(cam, 0x00, sizeof(*cam));
  kref_init(&cam->kref);

  cam->interface = interface;
  cam->udev = dev;

  if (!(cam->control_buffer = kmalloc(0x100, GFP_KERNEL))) {
    err("kmalloc() failed");
    goto error;
  }
  memset(cam->control_buffer, 0, 0x100);
  if (!(cam->hashtab = kmalloc(HASHSIZE*sizeof(struct nlist *), GFP_KERNEL))) {
    err("kmalloc() failed");
    goto error;
  }
  memset(cam->hashtab, 0, HASHSIZE*sizeof(struct nlist *));

  if ((le16_to_cpu(dev->descriptor.idVendor) == USB_M560X_VENDOR_ID)
    && (le16_to_cpu(dev->descriptor.idProduct) == USB_M560X_PRODUCT_M5602)) {
    camera_name = "Ali M5602-based WebCam";
    info("Ali M5602-based WebCam found.");
  }
  else if ((le16_to_cpu(dev->descriptor.idVendor) == USB_M560X_VENDOR_ID)
    && (le16_to_cpu(dev->descriptor.idProduct) == USB_M560X_PRODUCT_M5603c)) {
    camera_name = "Ali M5603c-based WebCam";
    info("Ali M5603c-based WebCam found.");
  } else {
    err("Vendor/Product ID do not match known M506x WebCam devices.");
    retval = -ENODEV;
    goto error;
  }
  cam->name = camera_name;

  /* set up the endpoint information */
  /* use only the first bulk-in and bulk-out endpoints */
  iface_desc = interface->cur_altsetting;

  for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
    endpoint = &iface_desc->endpoint[i].desc;

    if (!cam->int_in_endpointAddr
        && ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN)
        && ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_INT)) {
        /* we found an int in endpoint */
      buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
      cam->int_in_size = buffer_size;
      cam->int_in_endpointAddr = endpoint->bEndpointAddress;
// 			cam->int_in_buffer = kmalloc(buffer_size, GFP_KERNEL);
// 			if (!cam->int_in_buffer) {
// 				err("Could not allocate int_in_buffer");
// 				goto error;
// 			}
    }

    if (!cam->isoc_in_endpointAddr
        && ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN)
        && ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_ISOC)) {
        /* we found an isoc in endpoint */
      buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
      cam->isoc_in_size = buffer_size;
      cam->isoc_in_endpointAddr = endpoint->bEndpointAddress;
//      cam->isoc_in_buffer = kmalloc(buffer_size, GFP_KERNEL);
// 			if (!cam->isoc_in_buffer) {
// 				err("Could not allocate isoc_in_buffer");
// 				goto error;
// 			}
        }
  }
  if (!(cam->int_in_endpointAddr && cam->isoc_in_endpointAddr)) {
    err("Could not find both int-in and isoc-in endpoints");
    goto error;
  }

  /* save our data pointer in this interface device */
  usb_set_intfdata(interface, cam);

  cam->vdev = video_device_alloc();
  if (!cam->vdev) {
    retval = -ENOMEM;
    goto error;
  }

//   memcpy(cam->vdev, &v4l_m560x_template, sizeof (v4l_m560x_template));
//   static struct video_device v4l_m560x_template = {
  cam->vdev->owner    = THIS_MODULE;
//     .name     = "Ali M560x-based USB Camera",
    strcpy(cam->vdev->name, "Ali M560x-based USB Camera");
    cam->vdev->type     = VID_TYPE_CAPTURE;
    cam->vdev->hardware   = VID_HARDWARE_M560X;
    cam->vdev->fops     = &v4l_m560x_fops;
    cam->vdev->release  = video_device_release;
    cam->vdev->minor    = -1;
//   };
  if (cam->vdev)
    PDEBUG(DBG_TRACE, "cam->vdev exists");
//   cam->vdev->priv = cam;	// sort of a reverse mapping for those functions that get vdev only
//   cam->vdev->dev = &interface->dev;
  video_set_drvdata(cam->vdev,cam);
  
  memcpy (cam->vdev->name, cam->name, strlen (cam->name));
  init_waitqueue_head (&cam->wq);
  init_MUTEX (&cam->lock);
  init_MUTEX(&cam->buf_lock);
  for (i = 0; i < M560X_NUMFRAMES; i++) {
    cam->frame[i].framenum = i;
    cam->frame[i].width = 640;
    cam->frame[i].height = 480;
    init_waitqueue_head(&cam->frame[i].wq);
  }
  cam->grabframe = -1;
  wmb ();

  if (video_register_device(cam->vdev, VFL_TYPE_GRABBER, -1) == -1) {
    err("video_register_device failed");
      usb_set_intfdata(interface, NULL);
      video_device_release(cam->vdev);
      retval = -EIO;
      goto error;
  }

  info("m560x_probe - Ali_M560x WebCam driver is now controlling video device %d",cam->vdev->minor);

  m560x_create_sysfs_files(cam->vdev);

  return 0;

error:
  if (cam) {
// 		kref_put(&cam->kref, m560x_delete);
    kfree(cam);
  }
  return retval;
}

static void usb_m560x_disconnect(struct usb_interface *interface)
{
  struct m560x_camera *cam;
  int minor;

  cam = usb_get_intfdata(interface);
  minor = cam->vdev->minor;
 /* prevent m560x_open() from racing m560x_disconnect() */
 lock_kernel();

  usb_set_intfdata(interface, NULL);
  if (cam) {
    /* We don't want people trying to open up the device */
    if (cam->vdev) {
      m560x_remove_sysfs_files(cam->vdev);
      video_unregister_device(cam->vdev);
      cam->vdev = NULL;
    }
  }

  info("m560x_disconnect - Ali_M560x WebCam driver no longer controlling video device %d", minor);

  unlock_kernel();

  /* release resources */
  if (cam) {
    //usb_put_dev(cam->udev);
/*   if (cam->int_in_buffer)
     kfree (cam->int_in_buffer);
   if (cam->isoc_in_buffer)
     kfree (cam->isoc_in_buffer);*/
//     cam->udev = NULL;
    if (cam->control_buffer)
      kfree(cam->control_buffer);
    if (cam->hashtab) {
      clear_hashtab(cam->hashtab);
      kfree(cam->hashtab);
    }
    kfree (cam);
  }
 	
 /* decrement our usage count */
// 	kref_put(&cam->kref, m560x_delete);

	info("m560x_disconnect- M560x-based WebCam disconnected");;
}

static struct usb_driver usb_m560x_driver = {
	//.owner = THIS_MODULE,
	.name =	"m560x",
	.probe = usb_m560x_probe,
	.disconnect = usb_m560x_disconnect,
	.id_table = m560x_table,
};

static int __init usb_m560x_init(void)
{
	int result;

	info("usb_m560x_init - M560x-based WebCam driver startup");

	/* register this driver with the USB subsystem */
	result = usb_register(&usb_m560x_driver);
	if (result)
		err("usb_register failed. Error number %d", result);

	info(DRIVER_VERSION ":" DRIVER_DESC);
	
	return result;
}

static void __exit usb_m560x_exit(void)
{
	info("usb_m560x_exit - M560x-based WebCam driver shutdown");
	
	/* deregister this driver with the USB subsystem */
	usb_deregister(&usb_m560x_driver);
}


module_init (usb_m560x_init);
module_exit (usb_m560x_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
module_param(debug, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC (debug, "Debug enabled or not");
